// =================================================================================================
// Copyright 2002-2005 Adobe Systems Incorporated
// All Rights Reserved.
//
// NOTICE:  Adobe permits you to use, modify, and distribute this file in accordance with the terms
// of the Adobe license agreement accompanying it.
// =================================================================================================

#include "XMP_Environment.h"	// ! This must be the first include!
#include "XMP_Impl.hpp"

#include <cstring>

#if DEBUG
	#include <iostream>
#endif

using namespace std;

#if XMP_WinBuild
	#pragma warning ( disable : 4189 )	// local variable is initialized but not referenced
	#pragma warning ( disable : 4505 )	// unreferenced local function has been removed
#endif

// =================================================================================================

// *** This might be faster and use less memory as a state machine. A big advantage of building an
// *** XML tree though is easy lookahead during the recursive descent processing.

// *** It would be nice to give a line number or byte offset in the exception messages.


// 7 RDF/XML Grammar
//
// 7.1 Grammar summary
//
// 7.2.2 coreSyntaxTerms
//		rdf:RDF | rdf:ID | rdf:about | rdf:bagID | rdf:parseType | rdf:resource | rdf:nodeID | rdf:datatype
//
// 7.2.3 syntaxTerms
//		coreSyntaxTerms | rdf:Description | rdf:li
//
// 7.2.4 oldTerms
//		rdf:aboutEach | rdf:aboutEachPrefix
//
// 7.2.5 nodeElementURIs
//		anyURI - ( coreSyntaxTerms | rdf:li | oldTerms )
//
// 7.2.6 propertyElementURIs
//		anyURI - ( coreSyntaxTerms | rdf:Description | oldTerms )
//
// 7.2.7 propertyAttributeURIs
//		anyURI - ( coreSyntaxTerms | rdf:Description | rdf:li | oldTerms )
//
// 7.2.8 doc
//		root ( document-element == RDF, children == list ( RDF ) )
//
// 7.2.9 RDF
//		start-element ( URI == rdf:RDF, attributes == set() )
//		nodeElementList
//		end-element()
//
// 7.2.10 nodeElementList
//		ws* ( nodeElement ws* )*
//
// 7.2.11 nodeElement
//		start-element ( URI == nodeElementURIs,
//						attributes == set ( ( idAttr | nodeIdAttr | aboutAttr )?, bagIdAttr?, propertyAttr* ) )
//		propertyEltList
//		end-element()
//
// 7.2.12 ws
//		A text event matching white space defined by [XML] definition White Space Rule [3] in section Common Syntactic Constructs.
//
// 7.2.13 propertyEltList
//		ws* ( propertyElt ws* )*
//
// 7.2.14 propertyElt
//		resourcePropertyElt | literalPropertyElt | parseTypeLiteralPropertyElt |
//		parseTypeResourcePropertyElt | parseTypeCollectionPropertyElt | parseTypeOtherPropertyElt | emptyPropertyElt
//
// 7.2.15 resourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr? ) )
//		ws* nodeElement ws*
//		end-element()
//
// 7.2.16 literalPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, datatypeAttr?) )
//		text()
//		end-element()
//
// 7.2.17 parseTypeLiteralPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseLiteral ) )
//		literal
//		end-element()
//
// 7.2.18 parseTypeResourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseResource ) )
//		propertyEltList
//		end-element()
//
// 7.2.19 parseTypeCollectionPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseCollection ) )
//		nodeElementList
//		end-element()
//
// 7.2.20 parseTypeOtherPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseOther ) )
//		propertyEltList
//		end-element()
//
// 7.2.21 emptyPropertyElt
//		start-element ( URI == propertyElementURIs,
//						attributes == set ( idAttr?, ( resourceAttr | nodeIdAttr )?, bagIdAttr?, propertyAttr* ) )
//		end-element()
//
// 7.2.22 idAttr
//		attribute ( URI == rdf:ID, string-value == rdf-id )
//
// 7.2.23 nodeIdAttr
//		attribute ( URI == rdf:nodeID, string-value == rdf-id )
//
// 7.2.24 aboutAttr
//		attribute ( URI == rdf:about, string-value == URI-reference )
//
// 7.2.25 bagIdAttr
//		attribute ( URI == rdf:bagID, string-value == rdf-id )
//
// 7.2.26 propertyAttr
//		attribute ( URI == propertyAttributeURIs, string-value == anyString )
//
// 7.2.27 resourceAttr
//		attribute ( URI == rdf:resource, string-value == URI-reference )
//
// 7.2.28 datatypeAttr
//		attribute ( URI == rdf:datatype, string-value == URI-reference )
//
// 7.2.29 parseLiteral
//		attribute ( URI == rdf:parseType, string-value == "Literal")
//
// 7.2.30 parseResource
//		attribute ( URI == rdf:parseType, string-value == "Resource")
//
// 7.2.31 parseCollection
//		attribute ( URI == rdf:parseType, string-value == "Collection")
//
// 7.2.32 parseOther
//		attribute ( URI == rdf:parseType, string-value == anyString - ("Resource" | "Literal") )
//
// 7.2.33 URI-reference
//		An attribute string-value interpreted as an RDF URI Reference.
//
// 7.2.34 literal
//		Any XML element content that is allowed according to [XML] definition Content of Elements Rule [43] content
//		in section 3.1 Start-Tags, End-Tags, and Empty-Element Tags.
//
// 7.2.35 rdf-id
//		An attribute string-value matching any legal [XML-NS] token NCName.


// =================================================================================================
// Primary Parsing Functions
// =========================
//
// Each of these is responsible for recognizing an RDF syntax production and adding the appropriate
// structure to the XMP tree. They simply return for success, failures will throw an exception.

static void
RDF_RDF ( XMP_Node * xmpTree, const XML_Node & xmlNode );

static void
RDF_NodeElementList ( XMP_Node * xmpParent, const XML_Node & xmlParent, bool isTopLevel );

static void
RDF_NodeElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_NodeElementAttrs ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_PropertyElementList ( XMP_Node * xmpParent, const XML_Node & xmlParent, bool isTopLevel );
enum { kIsTopLevel = true, kNotTopLevel = false };

static void
RDF_PropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_ResourcePropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_LiteralPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_ParseTypeLiteralPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_ParseTypeResourcePropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_ParseTypeCollectionPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_ParseTypeOtherPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );

static void
RDF_EmptyPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel );


// =================================================================================================

typedef XMP_Uns8 RDFTermKind;

enum {
	kRDFTerm_Other				=  0,
	kRDFTerm_RDF				=  1,	// Start of coreSyntaxTerms.
	kRDFTerm_ID					=  2,
	kRDFTerm_about				=  3,
	kRDFTerm_bagID				=  4,
	kRDFTerm_parseType			=  5,
	kRDFTerm_resource			=  6,
	kRDFTerm_nodeID				=  7,
	kRDFTerm_datatype			=  8,	// End of coreSyntaxTerms.
	kRDFTerm_Description		=  9,	// Start of additions for syntaxTerms.
	kRDFTerm_li					= 10,	// End of of additions for syntaxTerms.
	kRDFTerm_aboutEach			= 11,	// Start of oldTerms.
	kRDFTerm_aboutEachPrefix	= 12	// End of oldTerms.
};

enum {
	kRDFMask_Other				= 1 << kRDFTerm_Other,
	kRDFMask_RDF				= 1 << kRDFTerm_RDF,
	kRDFMask_ID					= 1 << kRDFTerm_ID,
	kRDFMask_about				= 1 << kRDFTerm_about,
	kRDFMask_bagID				= 1 << kRDFTerm_bagID,
	kRDFMask_parseType			= 1 << kRDFTerm_parseType,
	kRDFMask_resource			= 1 << kRDFTerm_resource,
	kRDFMask_nodeID				= 1 << kRDFTerm_nodeID,
	kRDFMask_datatype			= 1 << kRDFTerm_datatype,
	kRDFMask_Description		= 1 << kRDFTerm_Description,
	kRDFMask_li					= 1 << kRDFTerm_li,
	kRDFMask_aboutEach			= 1 << kRDFTerm_aboutEach,
	kRDFMask_aboutEachPrefix	= 1 << kRDFTerm_aboutEachPrefix
};

enum {
	kRDF_HasValueElem = 0x10000000UL	// ! Contains rdf:value child. Must fit within kXMP_ImplReservedMask!
};

// -------------------------------------------------------------------------------------------------
// GetRDFTermKind
// --------------

static RDFTermKind
GetRDFTermKind ( const XMP_VarString & name )
{
	RDFTermKind term = kRDFTerm_Other;

	// Arranged to hopefully minimize the parse time for large XMP.

	if ( (name.size() > 4) && (strncmp ( name.c_str(), "rdf:", 4 ) == 0) ) {

		if ( name == "rdf:li" ) {
			term = kRDFTerm_li;
		} else if ( name == "rdf:parseType" ) {
			term = kRDFTerm_parseType;
		} else if ( name == "rdf:Description" ) {
			term = kRDFTerm_Description;
		} else if ( name == "rdf:about" ) {
			term = kRDFTerm_about;
		} else if ( name == "rdf:resource" ) {
			term = kRDFTerm_resource;
		} else if ( name == "rdf:RDF" ) {
			term = kRDFTerm_RDF;
		} else if ( name == "rdf:ID" ) {
			term = kRDFTerm_ID;
		} else if ( name == "rdf:bagID" ) {
			term = kRDFTerm_bagID;
		} else if ( name == "rdf:nodeID" ) {
			term = kRDFTerm_nodeID;
		} else if ( name == "rdf:datatype" ) {
			term = kRDFTerm_datatype;
		} else if ( name == "rdf:aboutEach" ) {
			term = kRDFTerm_aboutEach;
		} else if ( name == "rdf:aboutEachPrefix" ) {
			term = kRDFTerm_aboutEachPrefix;
		}

	}

	return term;

}	// GetRDFTermKind


// =================================================================================================


// -------------------------------------------------------------------------------------------------
// IsCoreSyntaxTerm
// ----------------
//
// 7.2.2 coreSyntaxTerms
//		rdf:RDF | rdf:ID | rdf:about | rdf:bagID | rdf:parseType | rdf:resource | rdf:nodeID | rdf:datatype

static bool
IsCoreSyntaxTerm ( RDFTermKind term )
{

	if 	( (kRDFTerm_RDF <= term) && (term <= kRDFTerm_datatype) ) return true;
	return false;

}	// IsCoreSyntaxTerm

// -------------------------------------------------------------------------------------------------
// IsSyntaxTerm
// ------------
//
// 7.2.3 syntaxTerms
//		coreSyntaxTerms | rdf:Description | rdf:li

static bool
IsSyntaxTerm ( RDFTermKind term )
{

	if 	( (kRDFTerm_RDF <= term) && (term <= kRDFTerm_li) ) return true;
	return false;

}	// IsSyntaxTerm

// -------------------------------------------------------------------------------------------------
// IsOldTerm
// ---------
//
// 7.2.4 oldTerms
//		rdf:aboutEach | rdf:aboutEachPrefix

static bool
IsOldTerm ( RDFTermKind term )
{

	if 	( (kRDFTerm_aboutEach <= term) && (term <= kRDFTerm_aboutEachPrefix) ) return true;
	return false;

}	// IsOldTerm

// -------------------------------------------------------------------------------------------------
// IsNodeElementName
// -----------------
//
// 7.2.5 nodeElementURIs
//		anyURI - ( coreSyntaxTerms | rdf:li | oldTerms )

static bool
IsNodeElementName ( RDFTermKind term )
{

	if 	( (term == kRDFTerm_li) || IsOldTerm ( term ) ) return false;
	return (! IsCoreSyntaxTerm ( term ));

}	// IsNodeElementName

// -------------------------------------------------------------------------------------------------
// IsPropertyElementName
// ---------------------
//
// 7.2.6 propertyElementURIs
//		anyURI - ( coreSyntaxTerms | rdf:Description | oldTerms )

static bool
IsPropertyElementName ( RDFTermKind term )
{

	if 	( (term == kRDFTerm_Description) || IsOldTerm ( term ) ) return false;
	return (! IsCoreSyntaxTerm ( term ));

}	// IsPropertyElementName

// -------------------------------------------------------------------------------------------------
// IsPropertyAttributeName
// -----------------------
//
// 7.2.7 propertyAttributeURIs
//		anyURI - ( coreSyntaxTerms | rdf:Description | rdf:li | oldTerms )

static bool
IsPropertyAttributeName ( RDFTermKind term )
{

	if 	( (term == kRDFTerm_Description) || (term == kRDFTerm_li) || IsOldTerm ( term ) ) return false;
	return (! IsCoreSyntaxTerm ( term ));

}	// IsPropertyAttributeName


// =================================================================================================
// AddChildNode
// ============

static XMP_Node *
AddChildNode ( XMP_Node * xmpParent, const XML_Node & xmlNode, const XMP_StringPtr value, bool isTopLevel )
{
	#if 0
		cout << "AddChildNode, parent = " << xmpParent->name << ", child = " << xmlNode.name;
		cout << ", value = \"" << value << '"';
		if ( isTopLevel ) cout << ", top level";
		cout << endl;
	#endif
	
	if ( xmlNode.ns.empty() ) {
		XMP_Throw ( "XML namespace required for all elements and attributes", kXMPErr_BadRDF );
	}
		
	XMP_StringPtr  childName    = xmlNode.name.c_str();
	const bool     isArrayItem  = (xmlNode.name == "rdf:li");
	const bool     isValueNode  = (xmlNode.name == "rdf:value");
	XMP_OptionBits childOptions = 0;
	
	if ( isTopLevel ) {

		// Lookup the schema node, adjust the XMP parent pointer.
		XMP_Assert ( xmpParent->parent == 0 );	// Incoming parent must be the tree root.
		XMP_Node * schemaNode = FindSchemaNode ( xmpParent, xmlNode.ns.c_str(), kXMP_CreateNodes );
		if ( schemaNode->options & kXMP_NewImplicitNode ) schemaNode->options ^= kXMP_NewImplicitNode;	// Clear the implicit node bit.
			// *** Should use "opt &= ~flag" (no conditional), need runtime check for proper 32 bit code.
		xmpParent = schemaNode;
		
		// If this is an alias set the isAlias flag in the node and the hasAliases flag in the tree.
		if ( sRegisteredAliasMap->find ( xmlNode.name ) != sRegisteredAliasMap->end() ) {
			childOptions |= kXMP_PropIsAlias;
			schemaNode->parent->options |= kXMP_PropHasAliases;
		}
		
	}

	// Make sure that this is not a duplicate of a named node.
	if ( ! (isArrayItem | isValueNode) ) {
		if ( FindChildNode ( xmpParent, childName, kXMP_ExistingOnly ) != 0 ) {
			XMP_Throw ( "Duplicate property or field node", kXMPErr_BadXMP );
		}
		
	}
	
	// Add the new child to the XMP parent node.
	XMP_Node * newChild = new XMP_Node ( xmpParent, childName, value, childOptions );
	if ( (! isValueNode) || xmpParent->children.empty() ) {
		 xmpParent->children.push_back ( newChild );
	} else {
		 xmpParent->children.insert ( xmpParent->children.begin(), newChild );
	}
	if ( isValueNode ) {
		if ( isTopLevel || (! (xmpParent->options & kXMP_PropValueIsStruct)) ) XMP_Throw ( "Misplaced rdf:value element", kXMPErr_BadRDF );
		xmpParent->options |= kRDF_HasValueElem;
	}
	
	if ( isArrayItem ) {
		if ( ! (xmpParent->options & kXMP_PropValueIsArray) ) XMP_Throw ( "Misplaced rdf:li element", kXMPErr_BadRDF );
		newChild->name = kXMP_ArrayItemName;
		#if 0	// *** XMP_DebugBuild
			newChild->_namePtr = newChild->name.c_str();
		#endif
	}
	
	return newChild;

}	// AddChildNode


// =================================================================================================
// AddQualifierNode
// ================

static XMP_Node *
AddQualifierNode ( XMP_Node * xmpParent, const XMP_VarString & name, const XMP_VarString & value )
{

	#if 0
		cout << "AddQualifierNode, parent = " << xmpParent->name << ", name = " << name;
		cout << ", value = \"" << value << '"' << endl;
	#endif
	
	const bool isLang = (name == "xml:lang");
	const bool isType = (name == "rdf:type");
	
	XMP_Node * newQual = 0;

		newQual = new XMP_Node ( xmpParent, name, value, kXMP_PropIsQualifier );

		if ( ! (isLang | isType) ) {
			xmpParent->qualifiers.push_back ( newQual );
		} else if ( isLang ) {
			if ( xmpParent->qualifiers.empty() ) {
				xmpParent->qualifiers.push_back ( newQual );
			} else {
				xmpParent->qualifiers.insert ( xmpParent->qualifiers.begin(), newQual );
			}
			xmpParent->options |= kXMP_PropHasLang;
		} else {
			XMP_Assert ( isType );
			if ( xmpParent->qualifiers.empty() ) {
				xmpParent->qualifiers.push_back ( newQual );
			} else {
				size_t offset = 0;
				if ( XMP_PropHasLang ( xmpParent->options ) ) offset = 1;
				xmpParent->qualifiers.insert ( xmpParent->qualifiers.begin()+offset, newQual );
			}
			xmpParent->options |= kXMP_PropHasType;
		}
		
		xmpParent->options |= kXMP_PropHasQualifiers;	// ! Don't set for the pxmp:compact.

	return newQual;

}	// AddQualifierNode


// =================================================================================================
// AddQualifierNode
// ================

static XMP_Node *
AddQualifierNode ( XMP_Node * xmpParent, const XML_Node & attr )
{
	if ( attr.ns.empty() ) {
		XMP_Throw ( "XML namespace required for all elements and attributes", kXMPErr_BadRDF );
	}
	
	return AddQualifierNode ( xmpParent, attr.name, attr.value );

}	// AddQualifierNode


// =================================================================================================
// FixupQualifiedNode
// ==================
//
// The parent is an RDF pseudo-struct containing an rdf:value field. Fix the XMP data model. The
// rdf:value node must be the first child, the other children are qualifiers. The form, value, and
// children of the rdf:value node are the real ones. The rdf:value node's qualifiers must be added
// to the others.
	
static void
FixupQualifiedNode ( XMP_Node * xmpParent )
{
	size_t qualNum, qualLim;

	XMP_Enforce ( (xmpParent->options & kXMP_PropValueIsStruct) && (! xmpParent->children.empty()) );

	XMP_Node * valueNode = xmpParent->children[0];
	XMP_Enforce ( valueNode->name == "rdf:value" );
	
	xmpParent->qualifiers.reserve ( xmpParent->qualifiers.size() + xmpParent->children.size() + valueNode->qualifiers.size() );
	
	// Move the qualifiers on the value node to the parent. Make sure an xml:lang qualifier stays at the front.
	
	qualNum = 0;
	qualLim = valueNode->qualifiers.size();
	
	if ( valueNode->options & kXMP_PropHasLang ) {
		if ( xmpParent->options & kXMP_PropHasLang ) XMP_Throw ( "Redundant xml:lang for rdf:value element", kXMPErr_BadXMP );
		XMP_Node * langQual = valueNode->qualifiers[0];
		langQual->parent = xmpParent;
		if ( xmpParent->qualifiers.empty() ) {
			xmpParent->qualifiers.push_back ( langQual );	// *** Should use utilities to add qual & set parent.
		} else {
			xmpParent->qualifiers.insert ( xmpParent->qualifiers.begin(), langQual );
		}
		qualNum = 1;	// Start the remaining copy after the xml:lang qualifier.
	}
	
	for ( ; qualNum != qualLim; ++qualNum ) {
		XMP_Node * currQual = valueNode->qualifiers[qualNum];
		currQual->parent = xmpParent;
		xmpParent->qualifiers.push_back ( currQual );
	}
	
	// Change the parent's other children into qualifiers.
	
	for ( qualNum = 1, qualLim = xmpParent->children.size(); qualNum != qualLim; ++qualNum ) {

		XMP_Node * currQual = xmpParent->children[qualNum];
	
			currQual->options |= kXMP_PropIsQualifier;
			currQual->parent = xmpParent;
			if ( xmpParent->qualifiers.empty() || (currQual->name != "xml:lang") ) {
				xmpParent->qualifiers.push_back ( currQual );
			} else {
				if ( xmpParent->options & kXMP_PropHasLang ) XMP_Throw ( "Redundant xml:lang for rdf:value element", kXMPErr_BadXMP );
				xmpParent->qualifiers.insert ( xmpParent->qualifiers.begin(), currQual );
				xmpParent->options |= kXMP_PropHasLang;	// *** Handle rdf:type also.
			}
		
	}
	
	if ( ! xmpParent->qualifiers.empty() ) xmpParent->options |= kXMP_PropHasQualifiers;
	
	// Move the options and value last, other checks need the parent's original options. Move the
	// value node's children to be the parent's children.
	
	XMP_Assert ( xmpParent->options & (kXMP_PropValueIsStruct | kRDF_HasValueElem) );
	xmpParent->options &= ~ (kXMP_PropValueIsStruct | kRDF_HasValueElem);
	xmpParent->options |= valueNode->options;
	
	xmpParent->value.swap ( valueNode->value );
	xmpParent->children.swap ( valueNode->children );
	
	for ( size_t childNum = 0, childLim = xmpParent->children.size(); childNum != childLim; ++childNum ) {
		XMP_Node * currChild = xmpParent->children[childNum];
		currChild->parent = xmpParent;
	}

	#if 0	// *** XMP_DebugBuild
		xmpParent->_valuePtr = xmpParent->value.c_str();
	#endif
	
	// Clear the value node's children and qualifier arrays and delete the now useless value node.
	// The destructor will delete remaining children and qualifiers, which we do not want to happen!
	
	valueNode->children.clear();
	valueNode->qualifiers.clear();
	delete valueNode;
	
}	// FixupQualifiedNode


// =================================================================================================
// ProcessRDF
// ==========
//
// Parse the XML tree of the RDF and build the corresponding XMP tree.

// *** Throw an exception if no XMP is found? By option?
// *** Do parsing exceptions cause the partial tree to be deleted?

void ProcessRDF ( XMP_Node * xmpTree, const XML_Node & rdfNode, XMP_OptionBits options )
{
	IgnoreParam(options);
	
	RDF_RDF ( xmpTree, rdfNode );

}	// ProcessRDF


// =================================================================================================
// RDF_RDF
// =======
//
// 7.2.9 RDF
//		start-element ( URI == rdf:RDF, attributes == set() )
//		nodeElementList
//		end-element()
//
// The top level rdf:RDF node. It can only have xmlns attributes, which have already been removed
// during construction of the XML tree.

static void
RDF_RDF ( XMP_Node * xmpTree, const XML_Node & xmlNode )
{

	if ( ! xmlNode.attrs.empty() ) XMP_Throw ( "Invalid attributes of rdf:RDF element", kXMPErr_BadRDF );
	RDF_NodeElementList ( xmpTree, xmlNode, kIsTopLevel );

}	// RDF_RDF


// =================================================================================================
// RDF_NodeElementList
// ===================
//
// 7.2.10 nodeElementList
//		ws* ( nodeElement ws* )*

static void
RDF_NodeElementList ( XMP_Node * xmpParent, const XML_Node & xmlParent, bool isTopLevel )
{
	XMP_Assert ( isTopLevel );
	
	XML_cNodePos currChild = xmlParent.content.begin();	// *** Change these loops to the indexed pattern.
	XML_cNodePos endChild  = xmlParent.content.end();

	for ( ; currChild != endChild; ++currChild ) {
		if ( IsWhitespaceNode ( **currChild ) ) continue;
		RDF_NodeElement ( xmpParent, **currChild, isTopLevel );
	}

}	// RDF_NodeElementList


// =================================================================================================
// RDF_NodeElement
// ===============
//
// 7.2.5 nodeElementURIs
//		anyURI - ( coreSyntaxTerms | rdf:li | oldTerms )
//
// 7.2.11 nodeElement
//		start-element ( URI == nodeElementURIs,
//						attributes == set ( ( idAttr | nodeIdAttr | aboutAttr )?, bagIdAttr?, propertyAttr* ) )
//		propertyEltList
//		end-element()
//
// A node element URI is rdf:Description or anything that is not an RDF term.

static void
RDF_NodeElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	RDFTermKind nodeTerm = GetRDFTermKind ( xmlNode.name );
	if ( (nodeTerm != kRDFTerm_Description) && (nodeTerm != kRDFTerm_Other) ) {
		XMP_Throw ( "Node element must be rdf:Description or typedNode", kXMPErr_BadRDF );
	}

	if ( isTopLevel && (nodeTerm == kRDFTerm_Other) ) {
		XMP_Throw ( "Top level typedNode not allowed", kXMPErr_BadXMP );
	} else {
		RDF_NodeElementAttrs ( xmpParent, xmlNode, isTopLevel );
		RDF_PropertyElementList ( xmpParent, xmlNode, isTopLevel );
	}

}	// RDF_NodeElement


// =================================================================================================
// RDF_NodeElementAttrs
// ====================
//
// 7.2.7 propertyAttributeURIs
//		anyURI - ( coreSyntaxTerms | rdf:Description | rdf:li | oldTerms )
//
// 7.2.11 nodeElement
//		start-element ( URI == nodeElementURIs,
//						attributes == set ( ( idAttr | nodeIdAttr | aboutAttr )?, bagIdAttr?, propertyAttr* ) )
//		propertyEltList
//		end-element()
//
// Process the attribute list for an RDF node element. A property attribute URI is anything other
// than an RDF term. The rdf:ID, rdf:nodeID, and rdf:bagID attributes are simply ignored, as are
// rdf:about attributes on inner nodes.

static const XMP_OptionBits kExclusiveAttrMask = (kRDFMask_ID | kRDFMask_nodeID | kRDFMask_about);

static void
RDF_NodeElementAttrs ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	XMP_OptionBits exclusiveAttrs = 0;	// Used to detect attributes that are mutually exclusive.

	XML_cNodePos currAttr = xmlNode.attrs.begin();
	XML_cNodePos endAttr  = xmlNode.attrs.end();

	for ( ; currAttr != endAttr; ++currAttr ) {

		RDFTermKind attrTerm = GetRDFTermKind ( (*currAttr)->name );

		switch ( attrTerm ) {

			case kRDFTerm_ID     :
			case kRDFTerm_nodeID :
			case kRDFTerm_about  :

				if ( exclusiveAttrs & kExclusiveAttrMask ) XMP_Throw ( "Mutally exclusive about, ID, nodeID attributes", kXMPErr_BadRDF );
				exclusiveAttrs |= (1 << attrTerm);

				if ( isTopLevel && (attrTerm == kRDFTerm_about) ) {
					// This is the rdf:about attribute on a top level node. Set the XMP tree name if
					// it doesn't have a name yet. Make sure this name matches the XMP tree name.
					XMP_Assert ( xmpParent->parent == 0 );	// Must be the tree root node.
					if ( ! xmpParent->name.empty() ) {
						if ( xmpParent->name != (*currAttr)->value ) XMP_Throw ( "Mismatched top level rdf:about values", kXMPErr_BadXMP );
					} else {
						xmpParent->name = (*currAttr)->value;
						#if 0	// *** XMP_DebugBuild
							xmpParent->_namePtr = xmpParent->name.c_str();
						#endif
					}
				}

				break;

			case kRDFTerm_bagID :
				// ! Don't have to check that rdf:bagID only occurs once, that is true for any XML attribute.
				break;

			case kRDFTerm_Other :
				AddChildNode ( xmpParent, **currAttr, (*currAttr)->value.c_str(), isTopLevel );
				break;

			default :
				XMP_Throw ( "Unrecognized nodeElement attribute", kXMPErr_BadRDF );

		}

	}

}	// RDF_NodeElementAttrs


// =================================================================================================
// RDF_PropertyElementList
// =======================
//
// 7.2.13 propertyEltList
//		ws* ( propertyElt ws* )*

static void
RDF_PropertyElementList ( XMP_Node * xmpParent, const XML_Node & xmlParent, bool isTopLevel )
{
	XML_cNodePos currChild = xmlParent.content.begin();
	XML_cNodePos endChild  = xmlParent.content.end();

	for ( ; currChild != endChild; ++currChild ) {
		if ( IsWhitespaceNode ( **currChild ) ) continue;
		if ( (*currChild)->kind != kElemNode ) {
			XMP_Throw ( "Expected property element node not found", kXMPErr_BadRDF );
		}
		RDF_PropertyElement ( xmpParent, **currChild, isTopLevel );
	}

}	// RDF_PropertyElementList


// =================================================================================================
// RDF_PropertyElement
// ===================
//
// 7.2.14 propertyElt
//		resourcePropertyElt | literalPropertyElt | parseTypeLiteralPropertyElt |
//		parseTypeResourcePropertyElt | parseTypeCollectionPropertyElt | parseTypeOtherPropertyElt | emptyPropertyElt
//
// 7.2.15 resourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr? ) )
//		ws* nodeElement ws*
//		end-element()
//
// 7.2.16 literalPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, datatypeAttr?) )
//		text()
//		end-element()
//
// 7.2.17 parseTypeLiteralPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseLiteral ) )
//		literal
//		end-element()
//
// 7.2.18 parseTypeResourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseResource ) )
//		propertyEltList
//		end-element()
//
// 7.2.19 parseTypeCollectionPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseCollection ) )
//		nodeElementList
//		end-element()
//
// 7.2.20 parseTypeOtherPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseOther ) )
//		propertyEltList
//		end-element()
//
// 7.2.21 emptyPropertyElt
//		start-element ( URI == propertyElementURIs,
//						attributes == set ( idAttr?, ( resourceAttr | nodeIdAttr )?, bagIdAttr?, propertyAttr* ) )
//		end-element()
//
// The various property element forms are not distinguished by the XML element name, but by their
// attributes for the most part. The exceptions are resourcePropertyElt and literalPropertyElt. They
// are distinguished by their XML element content.
//
// NOTE: The RDF syntax does not explicitly include the xml:lang attribute although it can appear in
// many of these. We have to allow for it in the attibute counts below.

static void
RDF_PropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	RDFTermKind nodeTerm = GetRDFTermKind ( xmlNode.name );
	if ( ! IsPropertyElementName ( nodeTerm ) ) XMP_Throw ( "Invalid property element name", kXMPErr_BadRDF );
	
	if ( xmlNode.attrs.size() > 3 ) {

		// Only an emptyPropertyElt can have more than 3 attributes.
		RDF_EmptyPropertyElement ( xmpParent, xmlNode, isTopLevel );

	} else {

		// Look through the attributes for one that isn't rdf:ID or xml:lang, it will usually tell
		// what we should be dealing with. The called routines must verify their specific syntax!

		XML_cNodePos currAttr = xmlNode.attrs.begin();
		XML_cNodePos endAttr  = xmlNode.attrs.end();
		XMP_VarString * attrName = 0;

		for ( ; currAttr != endAttr; ++currAttr ) {
			attrName = &((*currAttr)->name);
			if ( (*attrName != "xml:lang") && (*attrName != "rdf:ID") ) break;
		}

		if ( currAttr != endAttr ) {

			XMP_Assert ( attrName != 0 );
			XMP_VarString& attrValue = (*currAttr)->value;

			if ( *attrName == "rdf:datatype" ) {
				RDF_LiteralPropertyElement ( xmpParent, xmlNode, isTopLevel );
			} else if ( *attrName != "rdf:parseType" ) {
				RDF_EmptyPropertyElement ( xmpParent, xmlNode, isTopLevel );
			} else if ( attrValue == "Literal" ) {
				RDF_ParseTypeLiteralPropertyElement ( xmpParent, xmlNode, isTopLevel );
			} else if ( attrValue == "Resource" ) {
				RDF_ParseTypeResourcePropertyElement ( xmpParent, xmlNode, isTopLevel );
			} else if ( attrValue == "Collection" ) {
				RDF_ParseTypeCollectionPropertyElement ( xmpParent, xmlNode, isTopLevel );
			} else {
				RDF_ParseTypeOtherPropertyElement ( xmpParent, xmlNode, isTopLevel );
			}

		} else {

			// Only rdf:ID and xml:lang, could be a resourcePropertyElt, a literalPropertyElt, or an.
			// emptyPropertyElt. Look at the child XML nodes to decide which.

			if ( xmlNode.content.empty() ) {

				RDF_EmptyPropertyElement ( xmpParent, xmlNode, isTopLevel );

			} else {
			
				XML_cNodePos currChild = xmlNode.content.begin();
				XML_cNodePos endChild  = xmlNode.content.end();

				for ( ; currChild != endChild; ++currChild ) {
					if ( (*currChild)->kind != kCDataNode ) break;
				}
				
				if ( currChild == endChild ) {
					RDF_LiteralPropertyElement ( xmpParent, xmlNode, isTopLevel );
				} else {
					RDF_ResourcePropertyElement ( xmpParent, xmlNode, isTopLevel );
				}
			
			}

		}
		
	}

}	// RDF_PropertyElement


// =================================================================================================
// RDF_ResourcePropertyElement
// ===========================
//
// 7.2.15 resourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr? ) )
//		ws* nodeElement ws*
//		end-element()
//
// This handles structs using an rdf:Description node and arrays using rdf:Bag/Seq/Alt. It also
// catches qualified properties written with rdf:Description and rdf:value.

static void
RDF_ResourcePropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	if ( isTopLevel && (xmlNode.name == "iX:changes") ) return;	// Strip old "punchcard" chaff.
	
	XMP_Node * newCompound = AddChildNode ( xmpParent, xmlNode, "", isTopLevel );
	
	XML_cNodePos currAttr = xmlNode.attrs.begin();
	XML_cNodePos endAttr  = xmlNode.attrs.end();

	for ( ; currAttr != endAttr; ++currAttr ) {
		XMP_VarString& attrName = (*currAttr)->name;
		if ( (attrName != "rdf:ID") && (attrName != "xml:lang") ) {
			XMP_Throw ( "Invalid attribute for resource property element", kXMPErr_BadRDF );
		}
		AddQualifierNode ( newCompound, **currAttr );
	}
	
	XML_cNodePos currChild = xmlNode.content.begin();
	XML_cNodePos endChild  = xmlNode.content.end();

	for ( ; currChild != endChild; ++currChild ) {
		if ( ! IsWhitespaceNode ( **currChild ) ) break;
	}
	if ( currChild == endChild ) XMP_Throw ( "Missing child of resource property element", kXMPErr_BadRDF );
	if ( (*currChild)->kind != kElemNode ) XMP_Throw ( "Children of resource property element must be XML elements", kXMPErr_BadRDF );

	if ( (*currChild)->name == "rdf:Bag" ) {
		newCompound->options |= kXMP_PropValueIsArray;
	} else if ( (*currChild)->name == "rdf:Seq" ) {
		newCompound->options |= kXMP_PropValueIsArray | kXMP_PropArrayIsOrdered;
	} else if ( (*currChild)->name == "rdf:Alt" ) {
		newCompound->options |= kXMP_PropValueIsArray | kXMP_PropArrayIsOrdered | kXMP_PropArrayIsAlternate;
	} else {
		newCompound->options |= kXMP_PropValueIsStruct;
		if ( (*currChild)->name != "rdf:Description" ) {
			XMP_VarString typeName ( (*currChild)->ns );
			size_t        colonPos = (*currChild)->name.find_first_of(':');
			if ( colonPos == XMP_VarString::npos ) XMP_Throw ( "All XML elements must be in a namespace", kXMPErr_BadXMP );
			typeName.append ( (*currChild)->name, colonPos, XMP_VarString::npos );
			AddQualifierNode ( newCompound, XMP_VarString("rdf:type"), typeName );
		}
	}

	RDF_NodeElement ( newCompound, **currChild, kNotTopLevel );
	if ( newCompound->options & kRDF_HasValueElem ) {
		FixupQualifiedNode ( newCompound );
	} else if ( newCompound->options & kXMP_PropArrayIsAlternate ) {
		DetectAltText ( newCompound );
	}

	for ( ++currChild; currChild != endChild; ++currChild ) {
		if ( ! IsWhitespaceNode ( **currChild ) ) XMP_Throw ( "Invalid child of resource property element", kXMPErr_BadRDF );
	}

}	// RDF_ResourcePropertyElement


// =================================================================================================
// RDF_LiteralPropertyElement
// ==========================
//
// 7.2.16 literalPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, datatypeAttr?) )
//		text()
//		end-element()
//
// Add a leaf node with the text value and qualifiers for the attributes.

static void
RDF_LiteralPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	XMP_Node * newChild = AddChildNode ( xmpParent, xmlNode, "", isTopLevel );
	
	XML_cNodePos currAttr = xmlNode.attrs.begin();
	XML_cNodePos endAttr  = xmlNode.attrs.end();

	for ( ; currAttr != endAttr; ++currAttr ) {
		XMP_VarString & attrName = (*currAttr)->name;
		if ( (attrName != "rdf:ID") && (attrName != "rdf:datatype") && (attrName != "xml:lang") ) {
			XMP_Throw ( "Invalid attribute of literal property element", kXMPErr_BadRDF );
		}
		AddQualifierNode ( newChild, **currAttr );
	}
	
	XML_cNodePos currChild = xmlNode.content.begin();
	XML_cNodePos endChild  = xmlNode.content.end();
	size_t      textSize  = 0;

	for ( ; currChild != endChild; ++currChild ) {
		if ( (*currChild)->kind != kCDataNode ) XMP_Throw ( "Invalid child of literal property element", kXMPErr_BadRDF );
		textSize += (*currChild)->value.size();
	}
	
	newChild->value.reserve ( textSize );

	for ( currChild = xmlNode.content.begin(); currChild != endChild; ++currChild ) {
		newChild->value += (*currChild)->value;
	}

	#if 0	// *** XMP_DebugBuild
		newChild->_valuePtr = newChild->value.c_str();
	#endif
	
}	// RDF_LiteralPropertyElement


// =================================================================================================
// RDF_ParseTypeLiteralPropertyElement
// ===================================
//
// 7.2.17 parseTypeLiteralPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseLiteral ) )
//		literal
//		end-element()

static void
RDF_ParseTypeLiteralPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	IgnoreParam(xmpParent); IgnoreParam(xmlNode); IgnoreParam(isTopLevel); 
	
	XMP_Throw ( "ParseTypeLiteral property element not allowed", kXMPErr_BadXMP );

}	// RDF_ParseTypeLiteralPropertyElement


// =================================================================================================
// RDF_ParseTypeResourcePropertyElement
// ====================================
//
// 7.2.18 parseTypeResourcePropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseResource ) )
//		propertyEltList
//		end-element()
//
// Add a new struct node with a qualifier for the possible rdf:ID attribute. Then process the XML
// child nodes to get the struct fields.

static void
RDF_ParseTypeResourcePropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{

	XMP_Node * newStruct = AddChildNode ( xmpParent, xmlNode, "", isTopLevel );
	newStruct->options  |= kXMP_PropValueIsStruct;
	
	XML_cNodePos currAttr = xmlNode.attrs.begin();
	XML_cNodePos endAttr  = xmlNode.attrs.end();

	for ( ; currAttr != endAttr; ++currAttr ) {
		XMP_VarString& attrName = (*currAttr)->name;
		if ( attrName == "rdf:parseType" ) continue;	// ! The caller ensured the value is "Resource".
		if ( (attrName != "rdf:ID") && (attrName != "xml:lang") ) {
			XMP_Throw ( "Invalid attribute of ParseTypeResource property element", kXMPErr_BadRDF );
		}
		AddQualifierNode ( newStruct, **currAttr );
	}

	RDF_PropertyElementList ( newStruct, xmlNode, kNotTopLevel );

	if ( newStruct->options & kRDF_HasValueElem ) FixupQualifiedNode ( newStruct );
	
	// *** Need to look for arrays using rdf:Description and rdf:type.

}	// RDF_ParseTypeResourcePropertyElement


// =================================================================================================
// RDF_ParseTypeCollectionPropertyElement
// ======================================
//
// 7.2.19 parseTypeCollectionPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseCollection ) )
//		nodeElementList
//		end-element()

static void
RDF_ParseTypeCollectionPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	IgnoreParam(xmpParent); IgnoreParam(xmlNode); IgnoreParam(isTopLevel); 

	XMP_Throw ( "ParseTypeCollection property element not allowed", kXMPErr_BadXMP );

}	// RDF_ParseTypeCollectionPropertyElement


// =================================================================================================
// RDF_ParseTypeOtherPropertyElement
// =================================
//
// 7.2.20 parseTypeOtherPropertyElt
//		start-element ( URI == propertyElementURIs, attributes == set ( idAttr?, parseOther ) )
//		propertyEltList
//		end-element()

static void
RDF_ParseTypeOtherPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	IgnoreParam(xmpParent); IgnoreParam(xmlNode); IgnoreParam(isTopLevel); 

	XMP_Throw ( "ParseTypeOther property element not allowed", kXMPErr_BadXMP );

}	// RDF_ParseTypeOtherPropertyElement


// =================================================================================================
// RDF_EmptyPropertyElement
// ========================
//
// 7.2.21 emptyPropertyElt
//		start-element ( URI == propertyElementURIs,
//						attributes == set ( idAttr?, ( resourceAttr | nodeIdAttr )?, bagIdAttr?, propertyAttr* ) )
//		end-element()
//
// An empty property element can be encountered in 2 XMP data models, a simple property whose value
// is a URI, and a struct whose fields are all simple. The struct can also have an rdf:resource
// qualifier. The simple property case is a slightly different interpretation of RDF's formal model.
//
// The simple property looks like:
//		<ns:Prop rdf:resource="uri:prop"/>
// The XMP model is a simple property with the value "uri:prop" and the flag kXMP_PropValueIsURI.
//
// The structs without and with the qualifier look like:
//	<ns1:Stuct1 ns2:Field="value"/>
//	<ns1:Stuct2 ns2:Field="value" rdf:resource="uri:struct"/>
// Both structs have one field, ns2:Field with the value "value". Struct2 also has an rdf:resource
// qualifier with the value "uri:struct".
//
// In both of the above cases a child is added to the XMP parent node. The other possible attributes
// in the RDF end up being qualifiers.

// *** The various ID qualifiers are probably not serialized in this manner.

static void
RDF_EmptyPropertyElement ( XMP_Node * xmpParent, const XML_Node & xmlNode, bool isTopLevel )
{
	bool hasPropertyAttrs = false;
	bool hasResourceAttr  = false;
	bool hasNodeIDAttr    = false;
	
	// ! Because of implementation vagaries, the xmpParent is the tree root for top level properties.
	// ! The schema is found, created if necessary, by AddChildNode. Create an empty child, tweak it
	// ! as necessary when processing the attributes.
	
	if ( ! xmlNode.content.empty() ) XMP_Throw ( "Nested content not allowed with rdf:resource or property attributes", kXMPErr_BadRDF );
	
	XMP_Node * childNode = AddChildNode ( xmpParent, xmlNode, "", isTopLevel );
	
	XML_cNodePos currAttr = xmlNode.attrs.begin();
	XML_cNodePos endAttr  = xmlNode.attrs.end();

	for ( ; currAttr != endAttr; ++currAttr ) {

		RDFTermKind attrTerm = GetRDFTermKind ( (*currAttr)->name );

		switch ( attrTerm ) {

			case kRDFTerm_ID     :
			case kRDFTerm_bagID  :
				AddQualifierNode ( childNode, **currAttr );
				break;

			case kRDFTerm_nodeID :
				if ( hasResourceAttr ) XMP_Throw ( "Invalid attributes for empty property element", kXMPErr_BadRDF );
				AddQualifierNode ( childNode, **currAttr );
				hasNodeIDAttr = true;
				break;

			case kRDFTerm_resource :
				if ( hasNodeIDAttr ) XMP_Throw ( "Invalid attributes for empty property element", kXMPErr_BadRDF );
				if ( hasPropertyAttrs ) {
					AddQualifierNode ( childNode, **currAttr );
				} else {
					childNode->value = (*currAttr)->value;
					childNode->options |= kXMP_PropValueIsURI;
				}
				hasResourceAttr = true;
				break;

			case kRDFTerm_Other :
				if ( (*currAttr)->name == "xml:lang" ) {
					AddQualifierNode ( childNode, **currAttr );
				} else {
					if ( hasResourceAttr && (! hasPropertyAttrs) ) {
						// The first field, move the rdf:resource part from the value to a qualifier.
						AddQualifierNode ( childNode, "rdf:resource", childNode->value );
						childNode->value.erase();
						XMP_ClearOption ( childNode->options, kXMP_PropValueIsURI );
					}
					childNode->options |= kXMP_PropValueIsStruct;
					AddChildNode ( childNode, **currAttr, (*currAttr)->value.c_str(), false );
					hasPropertyAttrs = true;
				}
				break;

			default :
				XMP_Throw ( "Unrecognized attribute of empty property element", kXMPErr_BadRDF );
				break;

		}

	}

}	// RDF_EmptyPropertyElement


// =================================================================================================
